﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Speech.Synthesis;
using System.Runtime.InteropServices;

namespace voice
{
    class Program
    {
        // todo - just passthrough unless numlock/capslock/scrolllock detected?

        [DllImport("user32.dll", CharSet = CharSet.Auto, ExactSpelling = true, CallingConvention = CallingConvention.Winapi)]
        public static extern short GetKeyState(int keyCode);
        
        static void Help()
        {
            Console.WriteLine("voice.exe v0.2 by Eli Fulkerson 28 Dec 2015");
            Console.WriteLine("Just a thin command line wrapper around System.Speech.Synthesis.");
            Console.WriteLine("http://www.elifulkerson.com for updates");
            Console.WriteLine("");
            Console.WriteLine("Usage:  voice [-v VOLUME] [-r RATE] [-n NAME] [-m] [-f] [-l] [-d] [-?]");
            Console.WriteLine(" -v X   : Speak at volume X, where X is 0 to 100");
            Console.WriteLine(" -r X   : Speak at rate X, where X is -10 to 10");
            Console.WriteLine(" -n \"X\" : Speak using named voice X");
            Console.WriteLine(" -l     : List availabled named voices");
            Console.WriteLine(" -m     : Just use a male voice, if available");
            Console.WriteLine(" -f     : Just use a female voice, if available");
            Console.WriteLine(" -d     : (Optional) Indicate that arguments are done.");
            Console.WriteLine("          ... All text to the right of -d will be spoken even if it contains valid arguments.");
            Console.WriteLine(" -i     : Interactive mode");
            Console.WriteLine(" -t     : In pipe mode, output text to console as well as read it aloud");
            Console.WriteLine(" -s     : In pipe mode, copy to stderr and read aloud only when scroll lock is enabled");
            Console.WriteLine(" -?     : Help and version information");
            Console.WriteLine("");
            Console.WriteLine("Optional:");
            Console.WriteLine("For the sake of usability and/or self documentation, the *actual* arguments accepted are...");
            Console.WriteLine(" -v or /v or /volume or --volume");
            Console.WriteLine(" -r or /r or /rate or --rate");
            Console.WriteLine(" -n or /n or /name or --name");
            Console.WriteLine(" -l or /l or /list or --list");
            Console.WriteLine(" -m or /m or /male or --male");
            Console.WriteLine(" -f or /f or /female or --female");
            Console.WriteLine(" -d or /d or /done or --done");
            Console.WriteLine(" -i or /i or /interactive or --interactive");
            Console.WriteLine(" -t or /t or /tee or --tee");
            Console.WriteLine(" -s or /s or /scroll or --scroll");
            Console.WriteLine(" -? or --? or /? or /h or -h or /help or --help or -help or nothing at all");
            
        }

        static void Main(string[] args)
        {

            SpeechSynthesizer s = new SpeechSynthesizer();

            int Volume = s.Volume;
            int Rate = s.Rate;
            string line;
            bool mode_interactive = false;
            bool mode_tee = false;
            bool mode_scroll_inspect = false;

            int consumed = 0;  // the number of args that are used up - and thusly not the payload


            // This annoys me that I had to do this.
            bool keyAvailableHack = false;
            try
            {
                keyAvailableHack = System.Console.KeyAvailable;
            }
            catch
            {
                keyAvailableHack = true;
            }


            // help for no args unless we are piping
            if (args.Length == 0 && keyAvailableHack == false)
            {
                Help();
                Environment.Exit(0);
            }

            for (int i = 0; i < args.Length; i++)
            {
                if (args[i] == "-v" || args[i] == "--volume" || args[i] == "/v" || args[i] == "/volume")
                {
                    try
                    {
                        Volume = Convert.ToInt32(args[i + 1]);
                        if (Volume < 0 || Volume > 100)
                        {
                            Console.WriteLine("Volume must be between 0 and 100, inclusive.");
                            Environment.Exit(1);

                        }
                    }
                    catch
                    {
                        Console.WriteLine("Invalid volume specified");
                        Environment.Exit(1);
                    }

                    i++;
                    consumed += 2;
                    
                }

                if (args[i] == "-r" || args[i] == "--rate" || args[i] == "/r" || args[i] == "/rate")
                {
                    try
                    {
                        Rate = Convert.ToInt32(args[i + 1]);
                        if (Rate < -10 || Rate > 10)
                        {
                            Console.WriteLine("Rate must be between -10 and 10, inclusive.");
                            Environment.Exit(1);

                        }
                    }
                    catch
                    {
                        Console.WriteLine("Invalid rate specified");
                        Environment.Exit(1);
                    }

                    i++;
                    consumed += 2;

                }

                if (args[i] == "-m" || args[i] == "--male" || args[i] == "/m" || args[i] == "/male")
                {
                    s.SelectVoiceByHints(VoiceGender.Male);
                    consumed++;
                }

                if (args[i] == "-f" || args[i] == "--female" || args[i] == "/f" || args[i] == "/female")
                {
                    s.SelectVoiceByHints(VoiceGender.Female);
                    consumed++;
                }

                if (args[i] == "-n" || args[i] == "--name" || args[i] == "/n" || args[i] == "/name")
                {
                    try
                    {
                        s.SelectVoice(args[i + 1]);
                        
                    }
                    catch
                    {
                        Console.WriteLine("Invalid voice name specified.  Did you use quotes?  Try using --list to see available names");
                        Environment.Exit(1);
                    }
                    i++;
                    consumed+=2;
                }

                if (args[i] == "-l" || args[i] == "--list" || args[i] == "/l" || args[i] == "/list")
                {
                    //choose by name
                    foreach (InstalledVoice v in s.GetInstalledVoices())
                    {
                        Console.WriteLine("\"{0}\" - {1},{2},{3}", v.VoiceInfo.Name, v.VoiceInfo.Age, v.VoiceInfo.Gender, v.VoiceInfo.Culture);
                        //Console.WriteLine(v.VoiceInfo.Description);
                        //Console.WriteLine(v.VoiceInfo.Id);
                    }
                    Environment.Exit(0);
                }

                if (args[i] == "-?" || args[i] == "--?" || args[i] == "/?" || args[i] == "/h" || args[i] == "-h" || args[i] == "/help" || args[i] == "--help" || args[i] == "-help")
                {
                    //help and version and quit
                    Help();
                    Environment.Exit(0);
                }

                if (args[i] == "-i" || args[i] == "--interactive" || args[i] == "/i" || args[i] == "/interactive")
                {
                    mode_interactive = true;
                    consumed++;
                }

                if (args[i] == "-t" || args[i] == "--tee" || args[i] == "/t" || args[i] == "/tee")
                {
                    mode_tee = true;
                    consumed++;
                }

                if (args[i] == "-s" || args[i] == "--scroll" || args[i] == "/s" || args[i] == "/scroll")
                {
                    mode_scroll_inspect = true;
                    consumed++;
                }

                if (args[i] == "-d" || args[i] == "--done" || args[i] == "/d" || args[i] == "/done")
                {
                    // nothing after "done" gets parsed, in case there are actual codes in the speech
                    consumed++;
                    break;
                }
            }


            s.Volume = Volume;
            s.Rate = Rate;


            // "from the args" mode
            string tosay = String.Join(" ", args.Skip(consumed));
            if (tosay.Length > 0)
            {
                s.Speak(tosay);
            }

            // "from stdin, noninteractive" mode
            if (keyAvailableHack)
            {
             
                while (true)
                {
                    line = Console.ReadLine();
                    if (line == null)
                    {
                        break;
                    }
                    else
                    {

                        if (mode_scroll_inspect)
                        {
                            // check for scroll lock
                            if ((((ushort)GetKeyState(0x91)) & 0xffff) != 0)
                            {
                                Console.Error.WriteLine(line);
                                s.Speak(line);
                            }
                            Console.WriteLine(line);
                            continue;
                        } 

                        if (mode_tee)
                        {
                            Console.WriteLine(line);
                        }
                        s.Speak(line);
                    }
                }
            }
            
            // from stdin, interactive
            // "from stdin, noninteractive" mode


            if (mode_interactive)
            {
                Console.WriteLine("Interactive mode.  Control-c to quit.");
                while (true)
                {
                    line = Console.ReadLine();
                    if (line != null)
                    {

                        s.Speak(line);

                    }
                    else
                    {
                        break;
                    }
                }
            }
            
            
            
            Environment.Exit(0);
         
        }
    }
}
